%{ open Typing %}
%{ open Syntax %}
%{ open Preprocessing %}
%token <string> NONCE
%token <string> NONCE_STAR
%token <string> CONST
%token <string> KEY
%token <string> VAR
%token LEFT_PARAN
%token RIGHT_PARAN
%token DOT
%token COMMA
%right COMMA
%token COLON
%token EXCLAMATION
%token PIPE
%left PIPE
%token ZERO
%token NEW
%token IN
%token OUT
%token LET
%token ELSE
%token IF
%token THEN
%token MATCH
%token WITH
%token LL
%token LH
%token LS
%token KEYTYPE
%token EQUAL
%token EQUIV
%token ARROW
%token ENC
%token AENC
%token INL
%token INR
%token PK
%token VK
%token DEC
%token ADEC
%token SIGN
%token VERIFY
%token HASH
%token PROJ1
%token PROJ2
%token EOF
%token SEPARATOR
%token OR
%right OR
%token PLUS
%right PLUS
%token STAR
%right STAR
%token XL
%token XR
%token XLR
%token LEFT_BRACE
%token RIGHT_BRACE
%token UNDERSCORE

%nonassoc ITE
%nonassoc ELSE


%start <(Syntax.key_env * (Syntax.process * (Syntax.process * Syntax.process)) * (Syntax.process * (Syntax.process * Syntax.process))) option> top
%%


label:
  | LL {LL}
  | LH {LH}
  | LS {LS}

var:
  | x = VAR {x}

nonce :
  | n = NONCE {Nonce n}
  | n = NONCE_STAR {NonceStar n}

term:
  | LEFT_PARAN ; t = term RIGHT_PARAN {t}
  | n = nonce { VNonce (Error ($startpos,$endpos), n) }
  | x = var { VVar (Error ($startpos,$endpos), x) }
  | c = CONST { VConst (Error ($startpos,$endpos),c) }
  | k = KEY { VKey (Error ($startpos,$endpos),k) }
  | PK ; LEFT_PARAN ; k = KEY ; RIGHT_PARAN { VPKey (Error ($startpos,$endpos),(PKey k)) }
  | VK ; LEFT_PARAN ; k = KEY ; RIGHT_PARAN { VVerKey (Error ($startpos,$endpos),(VerKey k)) }
  | ENC ; LEFT_PARAN ; t = term ; COMMA ; k = KEY ; RIGHT_PARAN  { VSEnc (Error ($startpos,$endpos),t, k)}
  | SIGN ; LEFT_PARAN ; t = term ; COMMA ; k = KEY ; RIGHT_PARAN  { VSign (Error ($startpos,$endpos),t, k)}
  | AENC ; LEFT_PARAN ; t = term ; COMMA ; PK ; LEFT_PARAN ; k = KEY ; RIGHT_PARAN ; RIGHT_PARAN  { VAEnc (Error ($startpos,$endpos),t, PKey k)}
  | HASH ; LEFT_PARAN ; t = term ; RIGHT_PARAN  { VHash (Error ($startpos,$endpos),t)}
  | INL ; LEFT_PARAN ; t = term ; RIGHT_PARAN  { VInl (Error ($startpos,$endpos),t)}
  | INR ; LEFT_PARAN ; t = term ; RIGHT_PARAN  { VInr (Error ($startpos,$endpos),t)}
  | t1 = term ; COMMA ; t2 = term { VPair (Error ($startpos,$endpos),t1, t2) }

destructor : 
  | DEC ; LEFT_PARAN ; x = var ; COMMA ; k = KEY ; RIGHT_PARAN { DSDec (Error ($startpos,$endpos),x, k)}
  | ADEC ; LEFT_PARAN ; x = var ; COMMA ; k = KEY ; RIGHT_PARAN { DADec (Error ($startpos,$endpos),x, k)}
  | VERIFY ; LEFT_PARAN ; x = var ; COMMA ; VK ; LEFT_PARAN ; k = KEY ; RIGHT_PARAN ; RIGHT_PARAN { DVer (Error ($startpos,$endpos),x, VerKey k)}
  | PROJ1 ; LEFT_PARAN ; x = var ; RIGHT_PARAN ; { DProj1 (Error ($startpos,$endpos),x)}
  | PROJ2 ; LEFT_PARAN ; x = var ; RIGHT_PARAN ; { DProj2 (Error ($startpos,$endpos),x)}

proc:
  | ZERO { PZero (Error ($startpos,$endpos)) }
  | LEFT_PARAN ; p = proc1 ; RIGHT_PARAN {p}
  | NEW; n = nonce; COLON; l = label ; DOT ; p = proc { PNew (Error ($startpos,$endpos), n, l, p)}
  | OUT ; LEFT_PARAN ; t = term ; RIGHT_PARAN ; DOT ; p = proc { POut (Error ($startpos,$endpos), t, p)}
(*  | EXCLAMATION ; p = proc { PRepl (Error ($startpos,$endpos), p)} *)
  | IN ; LEFT_PARAN ; x = var ; RIGHT_PARAN ; DOT ; p = proc { PIn (Error ($startpos,$endpos), x, p)}
  | LET ; x = var ; EQUAL ; d = destructor ; IN ; p = proc ; ELSE ; q = proc { PLet (Error ($startpos,$endpos), x, d, p, q)}
  | LET ; x = var ; EQUAL ; d = destructor ; IN ; p = proc { PLet (Error ($startpos,$endpos), x, d, p, PZero(Error ($startpos,$endpos)))} %prec ITE
  | IF ; t1 = term ; EQUAL ; t2 = term ; THEN ; p = proc ; ELSE ; q = proc { PIf (Error ($startpos,$endpos), t1, t2, p, q)}
  | IF ; t1 = term ; EQUAL ; t2 = term ; THEN ; p = proc { PIf (Error ($startpos,$endpos), t1, t2, p, PZero(Error ($startpos,$endpos)))} %prec ITE
  | MATCH ; x = var ; WITH ; INL ; LEFT_PARAN ; y = var ; RIGHT_PARAN ; ARROW ; p = proc ; PIPE ; INR ; LEFT_PARAN ; z = var ; RIGHT_PARAN ; ARROW ; q = proc { PMatch (Error ($startpos,$endpos), x, y, p, z, q) }

proc1:
  | p = proc1 ; PIPE ; q = proc1 { PPar (Error ($startpos,$endpos),p,q)}
  | p = proc { p }

news:
  | {PZero ErTODO}
  | NEW; n = nonce; COLON; l = label ; DOT ; ns = news { PNew (Error ($startpos,$endpos), n, l, ns)}

proc_top : 
  | p = proc1 { p, (PZero (ErTODO),PZero (ErTODO))}
  | n = news ; EXCLAMATION ; LEFT_PARAN; p = proc1 ; RIGHT_PARAN {PZero (ErTODO), (n,p)}
  | n = news ; EXCLAMATION ; LEFT_PARAN ; p2 = proc1 ; RIGHT_PARAN ; PIPE ; p1 = proc1 {p1, (n,p2)}



typ:
  | l = label { TLabel l}
  | t = key_type {t}
  | t1 = typ ; PLUS ; t2 = typ {TSum (t1, t2)}
  | t1 = typ ; OR ; t2 = typ {TOr (t1, t2)}
  | t1 = typ ; STAR  ; t2 = typ {TPair (t1, t2)}
  | LEFT_BRACE; XL; EQUAL ; tl =  term ; COMMA ; XR ; EQUAL ; tr = term ; RIGHT_BRACE {TRefine (tl,tr)}
  | LEFT_BRACE; XLR; EQUAL ; t =  term ; RIGHT_BRACE {TRefine (t,t)}
  | LEFT_PARAN ; t = typ ; RIGHT_PARAN {t}
  | LEFT_PARAN ; t = typ ; RIGHT_PARAN ; UNDERSCORE ; k = KEY { TSEnc (t,k) }
  | LEFT_BRACE ; t = typ ; RIGHT_BRACE ; UNDERSCORE ; k = KEY { TAEnc (t,k) }

key_type :
  | KEYTYPE ; LEFT_PARAN ; l = label ; COMMA ; t = typ; RIGHT_PARAN { TKey (l,t) }

key_def : 
  | k = KEY ; COLON ; t = key_type { (k,t) }

key_env:
  | kd = key_def ; COMMA ; ke = key_env {fun x -> if x = fst kd then Some (snd kd) else ke x}
  | kd = key_def {fun x -> if x = fst kd then Some (snd kd) else None} 

top : 
  | EOF {None}
  | ke = key_env; SEPARATOR ; p = proc_top ; EQUIV ; q = proc_top ; EOF {Some (ke, p,q)}
  | ke = key_env; SEPARATOR ; p = proc_top ; EOF {Some (ke,p,p)}
  | p = proc_top ; EQUIV ; q = proc_top ; EOF {Some ((fun x -> None), p,q)}
  | p = proc_top ; EOF {Some ((fun x -> None), p,p)}
